"""
Provide a meta-model of something.

:constraint AAS-001:
    Some constraint
"""


class Something:
    """
    Represent something.

    See :constraintref:`AAS-NON-EXISTING`
    """


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
