"""
Do something.

    0: An unexpected block quote.
"""


class Something:
    """
    Represent something.

        1: An unexpected block quote.
    """

    something: str
    """
    Capture something.
    
        2: An unexpected block quote.
    """

    def __init__(self, something: str) -> None:
        self.something = something

    def do_something(self) -> None:
        """
        Do something.

            3: An unexpected block quote.
        """


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
