class Something:
    some_str: str

    def __init__(
        self,
        some_str: str,
    ) -> None:
        self.some_str = some_str


__version__ = "V198.4"

__xml_namespace__ = "https://dummy/198/4"
