class Something:
    """Represent something."""

    some_property: int
    """some property"""

    another_property: str

    yet_another_property: str
    """yet another property"""


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
