class Some_enum(Enum):
    """
    Represent something.

        Indented.
    """

    LITERAL = "Literal"
    """This is a literal."""

    LITERAL_WITHOUT_DOC = "LiteralWithoutDoc"


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"
